#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright 2018 2Cas
# Licensed under the MIT license (http://opensource.org/licenses/MIT)

import yaml, glob, os, errno, pathlib

from q2k.classes import *
from q2k.console import note_out, error_out
from q2k.globals import QMK_DIR, OUT_DIR, KBD_LIST, KB_INFO, keyplus_yaml_template


def special_file_dir(string):
    
    special_dir = ['handwired', 'converter', 'clueboard', 'lfkeyboards'] # Currently only lfkeyboards causes any issues, however it is good to be verbose here
    if string in special_dir:
        return True
    else:
        return False

def write_info(kb_info_yaml):
    templist = []
    keymaplist = []

    for fn in glob.glob(QMK_DIR+'keyboards/**/rules.mk', recursive=True):
        fn = fn[:-9]
        templist.append(fn)

    for fn in glob.glob(QMK_DIR+'keyboards/**/keymaps/**/keymap.c', recursive=True):
        fn = fn[:-9]
        if fn in templist:
            templist.remove(fn)
        fn = fn.replace(QMK_DIR+'keyboards/', '', 1)
        keymaplist.append(fn)

    for child in templist:
        p_path = str(pathlib.Path(child).parent)
        p_name = p_path.replace(QMK_DIR+'keyboards/', '', 1)

        if p_path not in templist and not special_file_dir(p_name):
            name = child.replace(QMK_DIR+'keyboards/', '', 1)
            # If in special dir list, then is part of the directory not keyboard
            if not special_file_dir(name):
               # This is a keyboard, not revision
               kb_obj = kb_info(name)
               kblibs = name.split('/')
               kb_obj.set_libs(kblibs) 
               KBD_LIST.append(kb_obj)   
        elif special_file_dir(p_name):
            name = child.replace(QMK_DIR+'keyboards/', '', 1)
            # This is a keyboard, not revision
            kb_obj = kb_info(name)
            kblibs = name.split('/')
            kb_obj.set_libs(kblibs)
            KBD_LIST.append(kb_obj)
        else:
            # This is a 'revision' of an existing keyboard
            rev = child.replace(p_path+'/', '')
            for kb_obj in KBD_LIST:
                if kb_obj.get_name() == p_name:
                    kb_obj.add_rev_list(rev)
                    break

    for kb_obj in KBD_LIST:
        if not kb_obj.get_rev_list():
            kb_obj.add_rev_list('n/a', True)

    for km_path in keymaplist:
        info_list = km_path.split('/keymaps/')
        kb_name = info_list[0]
        km_name = info_list[-1]
        namelist = kb_name.split('/')
        match = False
        for i in range(0, len(namelist)):
            if i > 0: 
                kb_name = '/'.join(namelist)

            for kb_obj in KBD_LIST:
                if kb_obj.get_name() == kb_name:
                    kb_obj.add_keymap_list(km_name)
                    match = True
                    break
            if match:
                break
            namelist.pop()

    if KBD_LIST:
        # Dump KBD_LIST to text file for faster processing in future
        dump_info(kb_info_yaml)
    else:
        error_out(['No keyboard information found', 'Is current directory a qmk root directory?', 'cd <qmk_dir>'])
        exit()

def dump_info(kb_info_yaml):

    kb_info_path = '/'.join(kb_info_yaml.split('/')[:-1])
    if not os.path.exists(kb_info_path):
        try:
            os.makedirs(kb_info_path)
        except OSError as e:
            if e.errno != errno.EEXIST and os.path.isdir(kb_info_path):
                raise
    try:
        with open(kb_info_yaml, 'w') as f:
            yaml.dump(KBD_LIST, f)
    except:
        error_out(['Failed to create '+kb_info_yaml])

def create_keyplus_yaml(kbc, printout=False):

    # Can't simply dump to yaml as we want to keep layout (array) as a human readable matrix (2D 'array').
    kb_n = kbc.get_name()
    rev_n = kbc.get_rev()
    keymap_n = kbc.get_keymap()

    rev = kbc.get_rev_info(rev_n)
    matrix = rev.get_matrix_pins()
    layers = rev.get_layout() 

    template_matrix = layers[0].get_template()
    name = '"'+kb_n+'_'+rev_n+'"'
    keymap = '"'+keymap_n+'"'
    if matrix:
        rows = str(matrix[0])
        cols = str(matrix[1])
    else:
        rows, cols = ''
   
    template = ''
    for i, row in enumerate(template_matrix):
        for col in row:
            template += (col+', ')
        if i+1 < len(template_matrix):
            template += '\n        '

    
    layout = ''
    for i, layer in enumerate(layers):
        layout += '      [ # layer '+str(i)+'\n        ['
        for row in layer.get_layout():
            layout += '\n          '
            for keycode in row:
               if len(keycode) < 4:
                   repeat = 4 - len(keycode)
                   for i in (range(repeat)):
                       keycode+=' '
               layout += keycode+', '
        layout +='\n        ]\n      ],\n'

    # Load Template
    output_yaml_info = keyplus_yaml_template
    output_yaml_info = output_yaml_info.replace('<KB_NAME>', name)
    output_yaml_info = output_yaml_info.replace('<LAYOUT_NAME>', keymap)                              
    output_yaml_info = output_yaml_info.replace('<ROWS>', rows)                                 
    output_yaml_info = output_yaml_info.replace('<COLS>', cols)
    output_yaml_info = output_yaml_info.replace('<MATRIX_MAP>', template)
    output_yaml_info = output_yaml_info.replace('<LAYOUT>', layout)

    kblibs = kbc.get_libs()
    if rev_n:
        path_list = kblibs + [ rev_n, keymap_n]
    else:
        path_list = kblibs + [keymap_n]
    output_path = '_'.join(path_list)
    output_yaml = OUT_DIR+output_path+'.yaml'
    if not os.path.exists(OUT_DIR):
        try:
            os.makedirs(OUT_DIR)
        except OSError as e:
            if e.errno != errno.EEXIST and os.path.isdir(output_path):
                raise
    try:
        with open(output_yaml, 'w') as f:
            f.write(output_yaml_info)
    except:
        error_out(['Failed to pipe output to '+output_yaml])
        exit()

    if printout:
        print(output_yaml_info)

    note_out(['SUCCESS! Output is in: '+output_yaml])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright 2018 2Cas
# Licensed under the MIT license (http://opensource.org/licenses/MIT)

import os, errno, subprocess

from q2k.classes import *
from q2k.globals import QMK_DIR, LOCAL_INCLUDES

def preproc(kb, kblibs, arg_list):
    qdir = QMK_DIR + 'keyboards/'

    # Setting up -I and custom define options
    cc = ['avr-gcc', '-E']
    kbdefine = 'KEYBOARD_'+'_'.join(kblibs)
    QMK_KEYBOARD_H = 'QMK_KEYBOARD_H=\"'+kb+'.h\"'
    libs = ['-D', kbdefine, '-D', QMK_KEYBOARD_H, '-I'+LOCAL_INCLUDES]
    path = qdir
    for kbl in kblibs:
        kbl += '/'
        path += kbl
        libs.append('-I'+path)
    argv = cc + libs + arg_list
    #print(' '.join(argv))
    try:
        output = subprocess.check_output(argv)
        return output
    except subprocess.CalledProcessError as e:
        err_num = e.returncode
        if err_num == 1:
            print(err_num)
            output = e.output
            return output
        else:
            print('*** Fatal compilation error')
            return

def preproc_header(kbc, path):
    qdir = QMK_DIR +'keyboards/'

    kb = kbc.get_name()
    kblibs = list(kbc.get_libs())
    rev = kbc.get_rev() 
    if rev:
        kblibs.append(rev)

    arg = [ '-D', 'QMK_KEYBOARD_CONFIG_H=\"config_common.h\"', '-dD', path ]
    
    output = str(preproc(kb, kblibs, arg)).replace('\\n', ' ')
    return output


def preproc_keymap(kbc):
    qdir = QMK_DIR +'keyboards/'

    kb = kbc.get_name()
    km = kbc.get_keymap()+'/keymap.c'
    kblibs = list(kbc.get_libs())
    rev = kbc.get_rev() 
    if rev != '':
        kblibs.append(rev)

    # KEYMAP
    keym = qdir+kb+'/keymaps/'+km
    if rev != '' and not os.path.isfile(keym):
        keym = qdir+kb+'/'+rev+'/keymaps/'+km

    # OUTPUT
    argkm = [keym]
    output = str(preproc(kb, kblibs, argkm)).replace('\\n', ' ')
    return output



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright 2018 2Cas
# Licensed under the MIT license (http://opensource.org/licenses/MIT)

import copy

from q2k.classes import *
from q2k.console import *

from q2k.globals import KEYP_FUNC_LIST, KEYP_KC_LIST

def convert_keymap(layers):

    l_list = []
    for ln in layers:
        name = ln.get_name()
        l_list.append(name)
    for l in layers:
        layer = l.get_keymap()
        for i, kc in enumerate(layer):
            if kc.endswith(')') and '(' in kc:
                layer[i] = func(kc, l_list)
            else:
                layer[i] = keycode(kc)
        l.set_keymap(layer)
    return layers

def func(qmk_func, layer_list):
    # Currently only handles layer switching functions
    if qmk_func in KEYP_FUNC_LIST.keys():
        keyp_func = KEYP_FUNC_LIST[qmk_func]
    else:

        try:
            brac_index = qmk_func.index('(')+1
            qfunc = qmk_func[:brac_index]
            func_target = qmk_func[brac_index:-1]
            if func_target in layer_list:
                layer = str(layer_list.index(func_target))
                keyp_func = KEYP_FUNC_LIST[qfunc]+layer
            else:
                bad_kc_out('FN','['+qmk_func+'] - set to trns')
                keyp_func = 'trns'
        except KeyError:
            bad_kc_out('FN','['+qmk_func+'] - set to trns')
            keyp_func = 'trns'
        except ValueError:
            bad_kc_out('FN','['+qmk_func+'] - set to trns')
            keyp_func = 'trns'

    return keyp_func


def keycode(qmk_kc):

    try:
        keyp_kc = KEYP_KC_LIST[qmk_kc]
    except KeyError:
        bad_kc_out('KC','['+qmk_kc+'] - set to trns')
        keyp_kc = 'trns'
    return keyp_kc


def build_layout_from_keymap(layers, layer_index=0):
    template_list = []

    keymap = layers[layer_index].get_keymap()
    col_limit = layers[layer_index].get_matrix_cols()
    matrix = []
    keymap_len = len(keymap)
    for i in range(keymap_len):
        if i % col_limit == 0:
            if i >= col_limit:
               matrix.append(row)
            row = []
            row.append(i)
        else:
            row.append(i)
    matrix.append(row)
    matrix_template = layout_template('!MATRIX LAYOUT')
    matrix_template.set_layout(matrix)

    template_list.append(matrix_template)
  
    return template_list            

def merge_layout_template(layers, templates, select=-1):

    if select >= 0:
        index = select
    if templates is None or len(templates) == 0:
        temp = build_layout_from_keymap(layers)
        merge_layout_template(layers, temp)
    if len(templates) == 1:
        index = 0
    elif select < 0:
        print('Select KEYMAP template:')
        for i, templ in enumerate(templates):
            print('['+str(i)+']\t'+ templ.get_name())

        while True:
            try:
                index = int(input())
            except ValueError:
                error_out(['Enter valid template number'])
                continue

            if index >= len(templates) or index < 0:
                error_out(['Enter valid template number'])
                continue
            else:
                break
    note_out(['Building with template: '+templates[index].get_name()])
    for x, l in enumerate(layers):    
        layout_name = templates[index].get_name()
        layout_template = templates[index].get_layout()
        layout = copy.deepcopy(layout_template)
        keycode_array = l.get_keymap()
        max_index = len(keycode_array)

        for i, rows in enumerate(layout_template):
            for j, ind in enumerate(rows):
                if ind < max_index:
                    layout[i][j] = keycode_array[ind]
                elif layout_name != '!MATRIX LAYOUT':
                    warning_out(['Corrupt/incompatible layout template or keymap', 
                        'Invalid array value: '+str(ind),
                        'Trying again with default matrix layout...'])
                    matrix_template = build_layout_from_keymap(layers, x)
                    merge_layout_template(layers, matrix_template)
                    exit()
                else:
                    error_out(['Corrupt/incompatible keymap',
                        'Invalid array value: '+str(ind)])
                    exit()
                    
        col_limit = l.get_matrix_cols()
        layout_template = convert_keyplus_matrix(layout_template, col_limit)
          
        l.set_layout(layout)
        l.set_template(layout_template)
        note_out(['Layer '+l.get_name()])
        """
        print(l.get_name())
        for row in layout: #l.get_keymap()
            print(row)

        for row in layout_template: #l.get_template()
            print(row)
        """

def convert_keyplus_matrix(matrix, col_limit):

     m = copy.deepcopy(matrix)
     for i,row in enumerate(matrix):
            for j,col in enumerate(row):
                int(col)
                r = col // col_limit
                c = col % col_limit
                m[i][j] = 'r'+str(r)+'c'+str(c)
     return m



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright 2018 2Cas
# Licensed under the MIT license (http://opensource.org/licenses/MIT)
import os

from q2k.classes import *
from q2k.console import note_out, error_out, bad_kc_out
from q2k.globals import QMK_DIR, OUT_KBF_DIR, KBD_LIST, KB_INFO, KBF_TRANSCODE, KBF_KC_LIST, KBF_FUNC_LIST

def check_keycode(kc, layer_list):

    if '(' in kc and ')' in kc:
        output = kbf_func(kc, layer_list)
    else:
        output = kbf_keycode(kc)
    return output

def kbf_func(func, layer_list):

    brac_index = func.index('(')+1
    qfunc = func[:brac_index]
    func_target = func[brac_index:-1]

    if func_target in layer_list:
        func_target = str(layer_list.index(func_target))

    if qfunc in KBF_TRANSCODE.keys():
        kbf_func = KBF_TRANSCODE[qfunc]+func_target+')'
        #bad_kc_out('FN','['+func+'] - set to '+kbf_func)
        return kbf_func
    elif qfunc in KBF_FUNC_LIST:
        return func
    else:
        kbf_func = 'KC_TRNS'
        bad_kc_out('FN','['+func+'] - set to KC_TRNS')
        return kbf_func

def kbf_keycode(kc):

    if kc in KBF_TRANSCODE.keys():
        kbf_kc = KBF_TRANSCODE[kc]
        #bad_kc_out('KC','['+kc+'] - set to '+kbf_kc)
        return kbf_kc
    elif kc in KBF_KC_LIST:
        return kc
    else:
        kbf_kc = 'KC_TRNS'
        bad_kc_out('KC','['+kc+'] - set to KC_TRNS')
        return kbf_kc

def keycode_func_parse_json(keycode, layer_list):

    func_list = []
    while '(' in keycode:
        split = keycode.split('(', 1)
        keycode = split[1][:-1]
        function = split[0]
        func_list.append(function)

    kc_data = ''
    for func in func_list:
        kc_data += '{"id": "'+func+'()","fields":['

    if keycode.isdigit():
        kc_data += keycode
    else:
        kc_data += '{"id": "'+check_keycode(keycode, layer_list)+'", "fields":[]}'

    for i in range(len(func_list)):
        kc_data += ']}'

    return kc_data


def create_kbfirmware_json(kbc, printout=False):

    kb_n = kbc.get_name()
    rev_n = kbc.get_rev()
    keymap_n = kbc.get_keymap()

    rev = kbc.get_rev_info(rev_n)
    matrix = rev.get_matrix_pins()
    layers = rev.get_layout() 

    template_matrix = layers[0].get_template()
    name = '"'+kb_n+'_'+rev_n+'"'
    keymap = '"'+keymap_n+'"'
    if matrix:
        m_rows = str(matrix[0]).replace("'", '"')
        m_cols = str(matrix[1]).replace("'", '"')
    else:
        m_rows, m_cols = '[]'

    json_out = '{"version":1,"keyboard":{"keys":['
    quantum = '"void matrix_init_user(void) {\\n}\\n\\nvoid matrix_scan_user(void) {\\n}\\n\\nbool process_record_user(uint16_t keycode, keyrecord_t *record) {\\n\\treturn true;\\n}"'
    id_list = []
    layer_names = []
    rmax = cmax = m_r_max = m_c_max = 0
    try:
        for n, layer in enumerate(layers):
            id = row_y = col_x = -1
            layer_names.append(layer.get_name())
            for i,row in enumerate(layer.get_layout()):
                row_y += 1
                for j, keycode in enumerate(row):
                    id += 1
                    col_x = col_x + 1
                    if n == 0:
                        keycodelist = [keycode]
                        id_list.append(keycodelist)
                    else:
                        id_list[id].append(keycode)

                    if n+1 == len(layers):
                        w = 1
                        h = 1
                        row_col = template_matrix[i][j].split('c')
                        r = row_col[0][1:]
                        c = row_col[-1]
                        if int(r) >= m_r_max:
                            m_r_max = int(r)+1
                        if int(c) >= m_c_max:
                            m_c_max = int(c)+1
                        kc_build = ''

                        for add in range(n+1, 16):
                            id_list[id].append('KC_TRNS')
                        for kc in id_list[id]:
                            kc = check_keycode(kc, layer_names)
                            kc_build += keycode_func_parse_json(kc, layer_names)+','
                        kc_build = kc_build[:-1]
                        json_out += ('{"id":'+str(id)+',"legend":"","state":{"x":'+str(col_x)+',"y":'+str(row_y)+
                                    ',"r":0,"rx":0,"ry":0,"w":'+w+',"h":'+h+'},"row":'+r+',"col":'+c+
                                    ',"keycodes":['+kc_build+']},')
                if col_x+1 > cmax:
                    cmax = col_x+1
                rmax = row_y+1
                col_x = -1
    except IndexError:
        error_out(['Layers have different layouts'])
        exit()
    json_out = json_out[:-1]
    json_out += '],"controller":1,"bounds":{"min":{"x":0,"y":0},"max":{"x":'+str(cmax)+',"y":'+str(rmax)+'}},'
    json_out += ('"rows":'+str(m_r_max)+',"cols":'+str(m_c_max)+',"pins":{"row":'+m_rows+',"col":'+m_cols+'},'+
                '"macros":{},"quantum":'+quantum+
                ',"settings":{"diodeDirection":0,"name":"","bootloaderSize":2,"rgbNum":0,"backlightLevels":3}}}')

    kblibs = kbc.get_libs()
    if rev_n:
        path_list = kblibs + [ rev_n, keymap_n]
    else:
        path_list = kblibs + [keymap_n]
    output_path = '_'.join(path_list)
    output_json = OUT_KBF_DIR+output_path+'.json'
    if not os.path.exists(OUT_KBF_DIR):
        try:
            os.makedirs(OUT_KBF_DIR)
        except OSError as e:
            if e.errno != errno.EEXIST and os.path.isdir(output_path):
                raise
    try:
        with open(output_json, 'w') as f:
            f.write(json_out)
    except:
        error_out(['Failed to pipe output to '+output_json])
        exit()

    if printout:
        print(json_out)

    note_out(['SUCCESS! Output is in: '+output_json, 'Note: Check WIRING + PINS + SETTINGS for unset values in KBFirmware'])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright 2018 2Cas
# Licensed under the MIT license (http://opensource.org/licenses/MIT)

from pkg_resources import Requirement, resource_filename

#----------------------------------------------
#MODIFY these to suit your liking BEFORE running pip install
KB_INFO = 'cache/kb_info.yaml'   # cached kb_info yaml
QMK_DIR = ''                     # QMK directory location
OUT_DIR = 'q2k_out/keyplus/'         # Default out directory
OUT_KBF_DIR = 'q2k_out/kbfirmware/'  # Default kbfirmware out directory
#----------------------------------------------

KBD_LIST = []
MCU_COMPAT = ['atmega8u2', 'atmega16u2', 'atmega32u2', 'atmega32u4', 'atmega32u6', 'at90usb646', 'at90usb646', 'at90usb647', 'at90usb1286', 'at90usb1287']
local_libs = resource_filename(Requirement.parse("q2k"),"q2k/lib/")
LOCAL_INCLUDES = local_libs+'/'

# QMK -> Keyplus Keycode Mapping
# for keycodes only, functions are in KEYP_FUNC_LIST
KEYP_KC_LIST = {
    "KC_NO"                 :  " no ", #- giorgio, 2018
    "KC_TRNS"               :  "____",
    "KC_TRANSPARENT"        :  "____", #"trans",
    "XXXXXXX"               :  " no ",
    "_______"               :  "____",
    "RESET"                 :  "reset",
    "MAGIC_HOST_NKRO"       :  "kro_n",                       # FORCE NKRO
    "MAGIC_UNHOST_NKRO"     :  "kro_6",                       # FORCE 6KRO
    "MAGIC_TOGGLE_NKRO"     :  "kro_auto",                    # Set Auto NKRO

    # Alphas
    "KC_GRV"                :  "'`' ",
    "KC_GRAVE"              :  "'`' ",  #"grave",             # Grave (~`)kro_6
    "KC_1"                  :  " 1  ",
    "KC_2"                  :  " 2  ",
    "KC_3"                  :  " 3  ",
    "KC_4"                  :  " 4  ",
    "KC_5"                  :  " 5  ",
    "KC_6"                  :  " 6  ",
    "KC_7"                  :  " 7  ",
    "KC_8"                  :  " 8  ",
    "KC_9"                  :  " 9  ",
    "KC_0"                  :  " 0  ",
    "KC_MINS"               :  " -  ",
    "KC_MINUS"              :  "'-' ", #"minus",               # Minus (-_)
    "KC_EQL"                :  "'=' ",
    "KC_EQUAL"              :  "'=' ", #"equal",               # Equal (=+)
    "KC_Q"                  :  " q  ",
    "KC_W"                  :  " w  ",
    "KC_E"                  :  " e  ",
    "KC_R"                  :  " r  ",
    "KC_T"                  :  " t  ",
    "KC_Y"                  :  " y  ",
    "KC_U"                  :  " u  ",
    "KC_I"                  :  " i  ",
    "KC_O"                  :  " o  ",
    "KC_P"                  :  " p  ",
    "KC_LBRC"               :  " [  ",
    "KC_LBRACKET"           :  " [  ", #"left_bracket",        # Left Bracket ([)
    "KC_RBRC"               :  " ]  ",
    "KC_RBRACKET"           :  " ]  ", #"right_bracket",       # Right Bracket (])
    "KC_BSLS"               :  "'\\' ", #"bsls",
    "KC_BSLASH"             :  "'\\' ",                        # Backslash (\|)
    "KC_A"                  :  " a  ",
    "KC_S"                  :  " s  ",
    "KC_D"                  :  " d  ",
    "KC_F"                  :  " f  ",
    "KC_G"                  :  " g  ",
    "KC_H"                  :  " h  ",
    "KC_J"                  :  " j  ",
    "KC_K"                  :  " k  ",
    "KC_L"                  :  " l  ",
    "KC_SCLN"               :  " ;  ",
    "KC_SCOLON"             :  " ;  ", #"semicolon",          # Semicolon (;:)
    "KC_QUOT"               :  " \"\' \"", 
    "KC_QUOTE"              :  " \"\' \"", #"quot",           # Quote ('")
    "KC_Z"                  :  " z  ",
    "KC_X"                  :  " x  ",
    "KC_C"                  :  " c  ",
    "KC_V"                  :  " v  ",
    "KC_B"                  :  " b  ",
    "KC_N"                  :  " n  ",
    "KC_M"                  :  " m  ",
    "KC_COMM"               :  "',' ",
    "KC_COMMA"              :  "',' ", #"comma",              # Comma (,<)
    "KC_DOT"                :  "'.' ",                        # Period (.>)
    "KC_SLSH"               :  "'/' ",
    "KC_SLASH"              :  "'/' ", #"forward_slash",      # Forward Slash (/?)
    "KC_SPC"                :  " spc",
    "KC_SPACE"              :  " spc", #"spacebar",           # Spacebar

    # Modifiers
    "KC_RCTL"               :  "rctl",
    "KC_RCTRL"              :  "rctl", #"rctrl",              # Right Ctrl
    "KC_LCTL"               :  "lctl",
    "KC_LCTRL"              :  "lctl", #"lctrl",              # Left Ctrl
    "KC_RSFT"               :  "rsft",
    "KC_RSHIFT"             :  "rsft", #"rshift",             # Right Shift
    "KC_LSFT"               :  "lsft",
    "KC_LSHIFT"             :  "lsft", #"lshift",             # Left Shift
    "KC_LGUI"               :  "lgui",
    "KC_LWIN"               :  "lgui", #"lwin",
    "KC_LCMD"               :  "lgui",                        # Left GUI/Win
    "KC_RGUI"               :  "rgui",
    "KC_RWIN"               :  "rgui", #"rwin",                    
    "KC_RCMD"               :  "rgui",                        # Right GUI/Win
    "KC_LALT"               :  "lalt",                        # Left Alt
    "KC_RALT"               :  "ralt",                        # Right Alt
    "KC_BSPC"               :  "bspc",
    "KC_BSPACE"             :  "bspc", #"backspace",          # Backspace
    "KC_ENT"                :  " ent",
    "KC_ENTER"              :  " ent", #"enter",              # Enter
    "KC_TAB"                :  " tab",                        # Tab
    "KC_CAPS"               :  "caps",
    "KC_CLCK"               :  "caps",
    "KC_CAPSLOCK"           :  "caps", #"caps_lock""kp_ent",, # Caps Lock
    "KC_RGHT"               :  "rght",
    "KC_RIGHT"              :  "rght", #"right",              # right arrow
    "KC_UP"                 :  " up ",
    "KC_DOWN"               :  "down",
    "KC_LEFT"               :  "left",
    # Function
    "KC_ESC"                :  " esc",
    "KC_ESCAPE"             :  " esc", #"escape",             # Escape
    "KC_F1"                 :  " f1 ",

    "KC_F2"                 :  " f2 ",
    "KC_F3"                 :  " f3 ",
    "KC_F4"                 :  " f4 ",
    "KC_F5"                 :  " f5 ",
    "KC_F6"                 :  " f6 ",
    "KC_F7"                 :  " f7 ",
    "KC_F8"                 :  " f8 ",
    "KC_F9"                 :  " f9 ",
    "KC_F10"                :  " f10",
    "KC_F11"                :  " f11",
    "KC_F12"                :  " f12",
    "KC_PSCR"               :  "pscr",
    "KC_PSCREEN"            :  "pscr", #"print_screen",      # Print Screen
    "KC_SLCK"               :  "slck",
    "KC_SCROLLLOCK"         :  "slck", #"scroll_lock",       # Scroll Lock
    "KC_PAUS"               :  "paus", 
    "KC_BRK"                :  "paus",
    "KC_PAUSE"              :  "paus", #"pause",             # Pause/Break
    "KC_INS"                :  " ins",
    "KC_INSERT"             :  " ins", #"insert",            # Insert
    "KC_DEL"                :  " del",
    "KC_DELETE"             :  " del", #"delete",            # Delete
    "KC_HOME"               :  "home",                       # Home
    "KC_END"                :  " end",                       # End
    "KC_PGUP"               :  "pgup",                       # Page Up
    "KC_PGDN"               :  "pgdn",
    "KC_PGDOWN"             :  "pgdn", #"page_down",         # Page Down
    "KC_F13"                :  " f13",
    "KC_F14"                :  " f14",
    "KC_F15"                :  " f15",
    "KC_F16"                :  " f16",
    "KC_F17"                :  " f17",
    "KC_F18"                :  " f18",
    "KC_F19"                :  " f19",
    "KC_F20"                :  " f20",
    "KC_F21"                :  " f21",
    "KC_F22"                :  " f22",
    "KC_F23"                :  " f23",
    "KC_F24"                :  " f24",
    # Shifted
    "KC_TILDE"              :  "'~' ",
    "KC_TILD"               :  "'~' ",
    "KC_EXCLAIM"            :  "'!' ",
    "KC_EXLM"               :  "'!' ",
    "KC_AT"                 :  "'@' ",
    "KC_HASH"               :  "'#' ",
    "KC_DOLLAR"             :  "'$' ",
    "KC_DLR"                :  "'$' ",
    "KC_PERCENT"            :  "'%' ",
    "KC_PERC"               :  "'%' ",
    "KC_CIRCUMFLEX"         :  "'^' ",
    "KC_CIRC"               :  "'^' ",
    "KC_AMPERSAND"          :  "'&' ",
    "KC_AMPR"               :  "'&' ",
    "KC_ASTERISK"           :  "'*' ",
    "KC_ASTR"               :  "'*' ",
    "KC_LEFT_PAREN"         :  "'(' ",
    "KC_LPRN"               :  "'(' ",
    "KC_RIGHT_PAREN"        :  "')' ",
    "KC_RPRN"               :  "')' ",
    "KC_UNDERSCORE"         :  "'_' ",
    "KC_UNDS"               :  "'_' ",
    "KC_PLUS"               :  "'+' ",
    "KC_LEFT_CURLY_BRACE"   :  "'{' ",
    "KC_LCBR"               :  "'{' ",
    "KC_RIGHT_CURLY_BRACE"  :  "'}' ",
    "KC_RCBR"               :  "'}' ",
    "KC_PIPE"               :  "'|' ",
    "KC_COLON"              :  "':' ",
    "KC_COLN"               :  "':' ",
    "KC_DOUBLE_QUOTE"       :  "\' \"\'",
    "KC_DQT"                :  "\' \"\'",
    "KC_DQUO"               :  "\' \"\'",
    "KC_LEFT_ANGLE_BRACKET" :  "'<' ",
    "KC_LT"                 :  "'<' ",
    "KC_LABK"               :  "'<' ",
    "KC_RIGHT_ANGLE_BRACKET":  "'>' ",
    "KC_GT"                 :  "'>' ",
    "KC_RABK"               :  "'>' ",
    "KC_QUESTION"           :  "'?' ",
    "KC_QUES"               :  "'?' ",

    # Numpad
    "KC_NLCK"               :  "nlck", 
    "KC_NUMLOCK"            :  "nlck", #"num_lock",        # Num Lock
    "KC_P1"                 :  "kp_1",
    "KC_KP_1"               :  "kp_1",                     # Numpad 1
    "KC_P2"                 :  "kp_2",   
    "KC_KP_2"               :  "kp_2",                     # Numpad 2
    "KC_P3"                 :  "kp_3",   
    "KC_KP_3"               :  "kp_3",                     # Numpad 3
    "KC_P4"                 :  "kp_4",   
    "KC_KP_4"               :  "kp_4",                     # Numpad 4
    "KC_P5"                 :  "kp_5",   
    "KC_KP_5"               :  "kp_5",                     # Numpad 5
    "KC_P6"                 :  "kp_6",   
    "KC_KP_6"               :  "kp_6",                     # Numpad 6
    "KC_P7"                 :  "kp_7",   
    "KC_KP_7"               :  "kp_7",                     # Numpad 7
    "KC_P8"                 :  "kp_8",   
    "KC_KP_8"               :  "kp_8",                     # Numpad 8
    "KC_P9"                 :  "kp_9",   
    "KC_KP_9"               :  "kp_9",                     # Numpad 9
    "KC_P0"                 :  "kp_0",   
    "KC_KP_0"               :  "kp_0",                     # Numpad 0
    "KC_PDOT"               :  "kp_.",
    "KC_KP_DOT"             :  "kp_.",                     # Numpad .
    "KC_PCMM"               :  "kp_,",
    "KC_KP_COMMA"           :  "kp_,",                     # Numpad ,
    "KC_PSLS"               :  "kp_/",
    "KC_KP_SLASH"           :  "kp_/",                     # Numpad /
    "KC_PAST"               :  "kp_*",                      
    "KC_KP_ASTERISK"        :  "kp_*",                     # Numpad *
    "KC_PMNS"               :  "kp_-", 
    "KC_KP_MINUS"           :  "kp_-",                     # Numpad -
    "KC_PPLS"               :  "kp_+",
    "KC_KP_PLUS"            :  "kp_+",                     # Numpad +
    "KC_PEQL"               :  "kp_=",
    "KC_KP_EQUAL"           :  "kp_=",                     # Numpad =
    "KC_PENT"               :  "kp_ent",
    "KC_KP_ENTER"           :  "kp_ent", #"kp_enter",      # Numpad Enter

    # Misc Functions
    "KC_APP"                :  " app",
    "KC_APPLICATION____"    :  "application",              # Application
    "KC_LCAP"               :  "locking_caps_lock",
    "KC_EXEC"               :  "execute",
    "KC_EXECUTE"            :  "execute",                  # Execute
    "KC_SLCT"               :  "select",
    "KC_SELECT"             :  "select",                   # Select____
    "KC_AGIN"               :  "again", 
    "KC_AGAIN"              :  "again",                    # Again
    "KC_MENU"               :  "hid_menu",                 # Menu
    "KC_UNDO"               :  "undo",                     # Undo
    "KC_CUT"                :  " cut",                      # Cut
    "KC_COPY"               :  "copy",                     # Copy
    "KC_PSTE"               :  "paste", 
    "KC_PASTE"              :  "paste",                    # Paste
    "KC_FIND"               :  "find",                     # Find
    "KC_ALT_ERASE"          :  "alternate_erase",          #KC_LEFT_ANGLE_BRACKET Alt Erase
    "KC_CANCEL"             :  "cancel",                   # Cancel
    "KC_SYSREQ"             :  "sys_req",                  # SYSREQ
    "KC_PRIOR"              :  "prior",                   
    "KC_SEPERATOR"          :  "separator",               
    "KC_RETURN"             :  "return",                   
    "KC_OUT"                :  " out",                      
    "KC_OPER"               :  "oper",                     
    "KC_CLEAR_AGAIN"        :  "clear_again",              
    "KC_CRSEL"              :  "crsel",
    "KC_EXSEL"              :  "exsel",
    "KC_STOP"               :  "stop",

    "KC_LOCKING_CAPS"       :  "locking_caps_lock",        # Locking Caps Lock
    "KC_LNUM"               :  "locking_num_lock",
    "KC_LOCKING_NUM"        :  "locking_num_lock",         # Locking Num Lock
    "KC_LSCR"               :  "locking_scroll_lock",
    "KC_LOCKING_SCROLL"     :  "locking_scroll_lock",      # Locking Scroll Lock
    "KC_ERAS"               :  "alternate_erase",
    "KC_ALT_ERASE"          :  "alternate_erase",          # Alternate Erase
    "KC_CLR"                :  "clear", 
    "KC_CLEAR"              :  "clear",                    # Clear
    "KC_NUHS"               :  "iso#",
    "KC_NONUS_HASH"         :  "iso_hash",                 # ISO hash (#~)
    "KC_NUBS"               :  "iso\\",
    "KC_NONUS_BSLASH"       :  "iso\\",                    # ISO Backslash (\|)
    "KC_ZKHK"               :  " `  ",                        # JIS Grave
    "KC_RO"                 :  "international_1",          
    "KC_INT1"               :  "international_1",          # JIS \|
    "KC_KANA"               :  "international_2",
    "KC_INT2"               :  "international_2",          # JIS Katakana/Hiragana
    "KC_JYEN"               :  "international_3",
    "KC_INT3"               :  "international_3",          # JIS Y
    "KC_HENK"               :  "international_4",
    "KC_INT4"               :  "international_4",          # JIS Henkan
    "KC_MHEN"               :  "international_5",
    "KC_INT5"               :  "international_5",          # JIS Muhenkan
    "KC_HAEN"               :  "lang_1",
    "KC_LANG1"              :  "lang_1",                   # KR Hangul/ENG
    "KC_HANJ"               :  "lang_2",
    "KC_LANG2"              :  "lang_2",                   # KR Hanja
 

    # Media Controls
    "KC__MUTE"              :  "mute",
    "KC_MUTE"               :  "mute",
    "KC_AUDIO_MUTE"         :  "mute", #"audio_mute",      # Audio Mute
    "KC_VOLU"               :  "volu",
    "KC__VOLUP"             :  "volu",
    "KC_AUDIO_VOL_UP"       :  "volu", #"audio_vol_up",    # Audio Vol. Up
    "KC_VOLD"               :  "vold", 
    "KC__VOLDOWN"           :  "vold",
    "KC_AUDIO_VOL_DOWN"     :  "vold", #"audio_vol_down",  # Audio Vol. Down
    "KC_MNXT"               :  "mnxt",
    "KC_MEDIA_NEXT_TRACK"   :  "mnxt", #"media_next_track",# Media Next Track
    "KC_MPRV"               :  "mprv",
    "KC_MEDIA_PREV_TRACK"   :  "mprv", #"media_prev_track",# Media Prev Track
    "KC_MFFD"               :  "mffd",
    "KC_MEDIA_FAST_FORWARD" :  "mffd",#"media_fast_forward",# Media Fast Forward
    "KC_MRWD"               :  "mrwd",
    "KC_MEDIA_REWIND"       :  "mrwd", #"media_rewind",    # Media Rewind
    "KC_MSTP"               :  "mstp",
    "KC_MEDIA_STOP"         :  "mstp", #"media_stop",      # Media Stop
    "KC_MPLY"               :  "mply",
    "KC_MEDIA_PLAY_PAUSE"   :  "mply", #"media_play_pause",# Media Play/Pause
    "KKC_DELETEC_MSEL"      :  "msel",
    "KC_MEDIA_SELECT"       :  "msel", #"media_select",    # Media Select
    "KC_EJCT"               :  "mjct",
    "KC_MEDIA_EJECT"        :  "mjct", #"media_eject",     # Media Eject

    # WWW
    "kro_6KC_MAIL"          :  "mail",                     # Mail
    "KC_CALC"               :  "calc",
    "KC_CALCULATOR"         :  "calculator",               # Calculator
    "KC_MYCM"               :  "comp",
    "KC_MY_COMPUTER"        :  "my_computer",              # My Computer
    "KC_WSCH"               :  "www_search",
    "KC_WWW_SEARCH"         :  "www_search",               # WWW Search
    "KC_WHOM"               :  "www_home", 
    "KC_WWW_HOME"           :  "www_home",                 # WWW Home
    "KC_WBAK"               :  "www_back", 
    "KC_WWW_BACK"           :  "www_back",                 # WWW Back
    "KC_WFWD"               :  "www_forward",
    "KC_WWW_FORWARD"        :  "www_forward",              # WWW Forward
    "KC_WSTP"               :  "www_stop",
    "KC_WWW_STOP"           :  "www_stop",                 # WWW Stop
    "KC_WREF"               :  "www_refresh",
    "KC_WWW_REFRESH"        :  "www_refresh",              # WWW Refresh
    "KC_WFAV"               :  "www_favourites",
    "KC_WWW_FAVORITES"      :  "www_favourites",           # WWW Favourites
    
    # Mousekey
    "KC_MS_U"               :  "ms_u",
    "KC_MS_UP"              :  "ms_u", #"mouse_up",        # Mouse Up
    "KC_MS_D"               :  "ms_d",
    "KC_MS_DOWN"            :  "ms_d", #"mouse_down",      # Mouse Down
    "KC_MS_L"               :  "ms_l", 
    "KC_MS_LEFT"            :  "ms_l", #"mouse_left",      # Mouse Left
    "KC_MS_R"               :  "ms_r", 
    "KC_MS_RIGHT"           :  "ms_r", #"mouse_right",     # Mouse Right
    "KC_BTN1"               :  "btn1",
    "KC_MS_BTN1"            :  "btn1", #"mouse_btn1",      # MOUSE1
    "KC_BTN2"               :  "btn2", 
    "KC_MS_BTN2"            :  "btn2", #"mouse_btn2",      # MOUSE2
    "KC_BTN3"               :  "btn3",
    "KC_MS_BTN3"            :  "btn3", #"mouse_btn3",      # MOUSE3
    "KC_BTN4"               :  "btn4",
    "KC_MS_BTN4"            :  "btn4", #"mouse_btn4",      # MOUSE4
    "KC_BTN5"               :  "btn5",
    "KC_MS_BTN5"            :  "btn5", #"mouse_btn5",      # MOUSE5
    "KC_WH_U"               :  "wh_u",
    "KC_MS_WH_UP"           :  "wh_u", #"mouse_wh_up",     # Mouse Wheel Up
    "KC_WH_D"               :  "wh_d",
    "KC_MS_WH_DOWN"         :  "wh_d", #"mouse_wh_down",   # Mouse Wheel Down
    "KC_WH_L"               :  "wh_l", 
    "KC_MS_WH_LEFT"         :  "wh_l", #"mouse_wh_left",   # Mouse Wheel Left
    "KC_WH_R"               :  "wh_r", 
    "KC_MS_WH_RIGHT"        :  "wh_r", #"mouse_wh_right",  # Mouse Wheel Right

    # System
    "KC_PWR"                :  "system_power",
    "KC_POWER"              :  "system_power",
    "KC_SYSTEM_POWER"       :  "system_power",             # Power
    "KC_SLEP"               :  "system_sleep",
    "KC_SYSTEM_SLEEP"       :  "system_sleep",             # Sleep
    "KC_WAKE"               :  "system_wake",
    "KC_SYSTEM_WAKE"        :  "system_wake",              # Wake
}

# QMK -> Keyplus functions
KEYP_FUNC_LIST = {
    "MO("                :  " l",
    "TG("                :  "toggle_l",
    "TT("                :  " l",            #Feature not fully implemented in K+
    "OSL("               :  "sticky_l",
    "DF("                :  "set_l",
    "OSM(KC_LSFT)"       :  "sticky_lshift",
    "OSM(KC_LSHIFT)"     :  "sticky_lshift",
    "OSM(KC_LCTL)"       :  "sticky_lctrl",
    "OSM(KC_LCTRL)"      :  "sticky_lctrl",
    "OSM(KC_LALT)"       :  "sticky_lalt",
    "OSM(KC_LGUI)"       :  "sticky_lgui",
    "OSM(KC_LWIN)"       :  "sticky_lgui",
    "OSM(KC_LCMD)"       :  "sticky_lgui",
    "OSM(KC_RSFT)"       :  "sticky_rshift",
    "OSM(KC_RSHIFT)"     :  "sticky_rshift",
    "OSM(KC_RCTL)"       :  "sticky_rctrl",
    "OSM(KC_RCTRL)"      :  "sticky_rctrl",
    "OSM(KC_RALT)"       :  "sticky_ralt",
    "OSM(KC_RGUI)"       :  "sticky_rgui",
    "OSM(KC_RWIN)"       :  "sticky_rgui",
    "OSM(KC_RCMD)"       :  "sticky_rgui", 
}


# QMK Keycodes currently not supported
keyp_missing = {
    # Setting Mouse Acceleration
    "KC_ACL0", "KC_MS_ACCEL0", "KC_ACL1", "KC_MS_ACCEL1", "KC_ACL2", "KC_MS_ACCEL2",
    # Thermal Printer
    "PRINT_ON", "PRINT_OFF",
    # Arr Gee Bee
    "RGB_TOG", "RGB_VAI", "RGB_VAD", "RGB_SAI", "RGB_MOD", "RGB_HUD", "RGB_SAD", "RGB_HUI",
    # Backlight
    "BL_STEP",
    # QMK Modifiers
    "LCTL(", "LALT(", "LGUI(", "LCMD(", "LWIN(", "RCTL(", "RSFT(", "RALT(", 
    "RGUI(", "RCMD(", "RWIN(", "KC_HYPR", "KC_MEH", "HYPR(", "MEH(", "LCAG(", 
    "ALTG(", "SGUI(", "SCMD(", "SWIN(", "LCA(", "LCTL_T(", "CTL_T(", "RCTL_T(",
    "LSFT_T(", "RSFT_T(", "LALT_T(", "ALT_T(", "RALT_T(", "ALGR_T(", "LGUI_T(",
    "LCMD_T(", "RWIN_T(", "GUI_T(", "RGUI_T(", "RCMD_T(", "RWIN_T(", "C_S_T(",
    "MEH_T(", "LCAG_T(" , "RCAG_T(", "ALL_T(", "SCMD_T(", "SWIN_T(", "LCA_T(",
}

# KBfirmware json compatability list
KBF_KC_LIST = [
    "KC_1", "KC_2", "KC_3", "KC_4", "KC_5", "KC_6", "KC_7",
    "KC_8", "KC_9", "KC_0",
    "KC_A", "KC_B", "KC_C", "KC_D", "KC_E", "KC_F", "KC_G",
    "KC_H", "KC_I", "KC_J", "KC_K", "KC_L", "KC_M", "KC_N",
    "KC_O", "KC_P", "KC_Q", "KC_R", "KC_S", "KC_T", "KC_U",
    "KC_V", "KC_W", "KC_X", "KC_Y", "KC_Z",
    "KC_NUHS", "KC_NUBS", "KC_MINS", "KC_EQL", "KC_LBRC", 
    "KC_RBRC", "KC_BSLS", "KC_SCLN", "KC_QUOT", "KC_GRV", 
    "KC_COMM", "KC_DOT",  "KC_SLSH", 
    "KC_ENT", "KC_ESC", "KC_BSPC", "KC_TAB", "KC_SPC", 
    "KC_CAPS", "KC_APP",
    "KC_LCTL", "KC_LSFT", "KC_LALT", "KC_LGUI", "KC_RCTL",
    "KC_RSFT", "KC_RALT", "KC_RGUI", "KC_TRNS", "KC_NO",
    "RESET",
    "KC_EXLM", "KC_AT", "KC_HASH", "KC_DLR", "KC_PERC",
    "KC_CIRC", "KC_AMPR", "KC_ASTR", "KC_LPRN", "KC_RPRN",
    "KC_UNDS", "KC_PLUS", "KC_LCBR", "KC_RCBR", "KC_PIPE",
    "KC_COLN", "KC_DQUO", "KC_TILD", "KC_LABK", "KC_RABK",
    "KC_QUES",
    "KC_F1", "KC_F2", "KC_F3", "KC_F4", "KC_F5", "KC_F6",
    "KC_F7", "KC_F8", "KC_F9", "KC_F10", "KC_F11", "KC_F12",
    "KC_F13", "KC_F14", "KC_F15", "KC_F16", "KC_F17",
    "KC_F18", "KC_F19", "KC_F20", "KC_F21", "KC_F22",
    "KC_F23", "KC_F24",
    "KC_PSCR", "KC_SLCK", "KC_PAUS",
    "KC_INS", "KC_DEL", "KC_HOME", "KC_END", "KC_PGUP", 
    "KC_PGDN", "KC_LEFT", "KC_DOWN", "KC_UP", "KC_RGHT",
    "KC_PWR", "KC_SLEP", "KC_WAKE", "KC_MUTE", "KC_VOLU",
    "KC_VOLD", "KC_MPLY", "KC_MSTP", "KC_MPRV", "KC_MNXT",
    "KC_RO", "KC_KANA", "KC_JYEN", "KC_HENK", "KC_MHEN",
    "KC_LOCKING_CAPS", "KC_LOCKING_NUM", "KC_LOCKING_SCROLL",
    "KC_APP", "KC_CALC", "KC_MYCM", "KC_WSCH", "KC_WHOM",
    "KC_WBAK", "KC_WFWD", "KC_WSTP", "KC_WREF", "KC_WFAV"
    "KC_MSEL",
    "KC_NLCK", "KC_PSLS", "KC_PAST", "KC_PMNS", "KC_PPLS",
    "KC_PDOT", "KC_PEQL", "KC_PENT", "KC_PCMM",
    "KC_P1", "KC_P2", "KC_P3", "KC_P4", "KC_P5", "KC_P6",
    "KC_P7", "KC_P8", "KC_P9", "KC_P0",
    "BL_TOGG", "BL_DEC", "BL_INC", "BL_STEP",
    "RGB_TOG", "RGB_MOD", "RGB_HUI", "RGB_HUD", "RGB_SAI",
    "RGB_SAD",  "RGB_VAI", "RGB_VAD",
    "KC_MS_U", "KC_MS_D", "KC_MS_L", "KC_MS_R", 
    "KC_BTN1", "KC_BTN2", "KC_BTN3", "KC_BTN4", "KC_BTN5",
    "KC_WH_U", "KC_WH_D", "KC_WH_L", "KC_WH_R", 
    "KC_ACL0", "KC_ACL1", "KC_ACL2",
]

KBF_FUNC_LIST = [
    "LCTL(", "LSFT(", "LALT(", "LGUI(", "RTL(", "RSFT(",
    "RALT(", "RGUI(", "HYPR(", "MEH(", "LCAG(", "ALTG(", 
    "LT(", "TO(", "MO(", "DF(", "TG(", "OSL(", "OSM(", 
    "MT(", "CTL_T(", "SFT_T(", "ALT_T(", "GUI_T(", 
    "C_S_T(", "MEH_T(", "LCAG_T(", "ALL_T(", "M("
]

KBF_TRANSCODE = {
    "S("                :    "LSFT(",
    "RCMD("             :    "RGUI(",
    "RWIN("             :    "RGUI(",
    "LCTL_T("           :    "CTL_T(",
    "RCTL_T("           :    "CTL_T(",
    "LSFT_T("           :    "SFT_T(",
    "RSFT_T("           :    "SFT_T(",
    "LALT_T("           :    "ALT_T(",
    "RALT_T("           :    "ALT_T(",
    "LGUI_T("           :    "GUI_T(",
    "LCMD_T("           :    "GUI_T(",
    "LWIN_T("           :    "GUI_T(",
    "RGUI_T("           :    "GUI_T(",
    "RCMD_T("           :    "GUI_T(",
    "RWIN_T("           :    "GUI_T(",
    "RCAG_T("           :    "LCAG_T(",
    "TT("               :    "MO(",
    "KC_ENTER"          :    "KC_ENT",
    "KC_ESCAPE"         :    "KC_ESC",
    "KC_BSCAPE"         :    "KC_BSPC",
    "KC_SPACE"          :    "KC_SPC",
    "KC_MINUS"          :    "KC_MINS",
    "KC_LBRACKET"       :    "KC_LBRC",
    "KC_RBRACKET"       :    "KC_RBRC",
    "KC_BSLASH"         :    "KC_BSLS",
    "KC_NONUS_HASH"     :    "KC_NUHS",
    "KC_NONUS_BSLASH"   :    "KC_NUBS",
    "KC_INT1"           :    "KC_RO",
    "KC_INT2"           :    "KC_KANA",
    "KC_INT3"           :    "KC_JYEN",
    "KC_INT4"           :    "KC_HENK", 
    "KC_INT5"           :    "KC_MHEN",
    "KC_SCOLON"         :    "KC_SCLN",
    "KC_QUOTE"          :    "KC_QUOT",
    "KC_GRAVE"          :    "KC_GRV",
    "KC_COMMA"          :    "KC_COMM",
    "KC_SLASH"          :    "KC_SLSH",
    "KC_CAPSLOCK"       :    "KC_CAPS",
    "KC_LCTRL"          :    "KC_LCTL",
    "KC_LSHIFT"         :    "KC_LSFT",
    "KC_LCMD"           :    "KC_LGUI",
    "KC_LWIN"           :    "KC_LGUI",
    "KC_RCTRL"          :    "KC_RCTL",
    "KC_RSHIFT"         :    "KC_RSFT",
    "KC_RCMD"           :    "KC_RGUI",
    "KC_RWIN"           :    "KC_RGUI",
    "KC_LCAP"           :    "KC_LOCKING_CAPS",
    "KC_LNUM"           :    "KC_LOCKING_NUM",
    "KC_LSCR"           :    "KC_LOCKING_SCROLL",
    "KC_PSCREEN"        :    "KC_PSCR",
    "KC_SCROLLLOCK"     :    "KC_SLCK",
    "KC_PAUSE"          :    "KC_PAUS",
    "KC_INSERT"         :    "KC_INS",
    "KC_DELETE"         :    "KC_DEL",
    "KC_PGDOWN"         :    "KC_PGDN",
    "KC_RIGHT"          :    "KC_RGHT",
    "KC_APPLICATION"    :    "KC_APP",
    "KC_SYSTEM_POWER"   :    "KC_PWR",
    "KC_SYSTEM_SLEEP"   :    "KC_SLEP",
    "KC_SYSTEM_WAKE"    :    "KC_WAKE",
    "KC_CALCULATOR"     :    "KC_CALC",
    "KC_MY_COMPUTER"    :    "KC_MYCM",
    "KC_WWW_SEARCH"     :    "KC_WSCH",
    "KC_WWW_HOME"       :    "KC_WHOM",
    "KC_WWW_BACK"       :    "KC_WBAK",
    "KC_WWW_FORWARD"    :    "KC_WFWD",
    "KC_WWW_STOP"       :    "KC_WSTP",
    "KC_WWW_REFRESH"    :    "KC_WREF",
    "KC_WWW_FAVORITES"  :    "KC_WFAV",
    "KC_AUDIO_MUTE"     :    "KC_MUTE",
    "KC_AUDIO_VOL_UP"   :    "KC_VOLU",
    "KC_AUDIO_VOL_DOWN" :    "KC_VOLD",
    "KC_MEDIA_NEXT_TRACK":   "KC_MNXT",
    "KC_MEDIA_PREV_TRACK":   "KC_MPRV",
    "KC_MEDIA_STOP"     :    "KC_MSTP",
    "KC_MEDIA_PLAY_PAUSE":   "KC_MPLY",
    "KC_MEDIA_SELECT"   :    "KC_MSEL",
    "KC_NUMLOCK"        :    "KC_NLCK",
    "KC_KP_SLASH"       :    "KC_PSLS",
    "KC_KP_ASTERISK"    :    "KC_PAST",
    "KC_KP_MINUS"       :    "KC_PMNS",
    "KC_KP_PLUS"        :    "KC_PPLS",
    "KC_KP_ENTER"       :    "KC_PENT",
    "KC_KP_1"           :    "KC_P1",
    "KC_KP_2"           :    "KC_P2",
    "KC_KP_3"           :    "KC_P3",
    "KC_KP_4"           :    "KC_P4",
    "KC_KP_5"           :    "KC_P5",
    "KC_KP_6"           :    "KC_P6",
    "KC_KP_7"           :    "KC_P7",
    "KC_KP_8"           :    "KC_P8",
    "KC_KP_9"           :    "KC_P9",
    "KC_KP_0"           :    "KC_P0",
    "KC_KP_DOT"         :    "KC_PDOT",
    "KC_KP_EQUAL"       :    "KC_PEQL",
    "KC_KP_COMMA"       :    "KC_PCMM",
    "KC_TRANSPARENT"    :    "KC_TRNS",
    "_______"           :    "KC_TRNS",
    "KC_MS_UP"          :    "KC_MS_U",
    "KC_MS_DOWN"        :    "KC_MS_D",
    "KC_MS_LEFT"        :    "KC_MS_L",
    "KC_MS_RIGHT"       :    "KC_MS_R",
    "KC_MS_BTN1"        :    "KC_BTN1",
    "KC_MS_BTN2"        :    "KC_BTN2",
    "KC_MS_BTN3"        :    "KC_BTN3",
    "KC_MS_BTN4"        :    "KC_BTN4",
    "KC_MS_BTN5"        :    "KC_BTN5",
    "KC_MS_WH_UP"       :    "KC_WH_U",
    "KC_MS_WH_DOWN"     :    "KC_WH_D",
    "KC_MS_WH_LEFT"     :    "KC_WH_L",
    "KC_MS_WH_RIGHT"    :    "KC_WH_R",
    "KC_MS_ACCEL0"      :    "KC_ACL0",
    "KC_MS_ACCEL1"      :    "KC_ACL1",
    "KC_MS_ACCEL2"      :    "KC_ACL2", 
    "RGB_MODE_FORWARD"  :    "RGB_MOD",
    "KC_TILDE"          :    "KC_TILD",
    "KC_EXCLAIM"        :    "KC_EXLM",
    "KC_DOLLAR"         :    "KC_DLR",
    "KC_PERCENT"        :    "KC_PERC",
    "KC_CIRCUMFLEX"     :    "KC_CIRC",
    "KC_AMPERSAND"      :    "KC_AMPR",
    "KC_ASTERISK"       :    "KC_ASTR",
    "KC_LEFT_PAREN"     :    "KC_LPRN",
    "KC_RIGHT_PAREN"    :    "KC_RPRN",
    "KC_UNDERSCORE"     :    "KC_UNDS",
    "KC_LEFT_CURLY_BRACE":   "KC_LCBR",
    "KC_RIGHT_CURLY_BRACE":  "KC_RCBR",
    "KC_COLON"          :    "KC_COLN",
    "KC_DOUBLE_QUOTE"   :    "KC_DQUO", 
    "KC_DQT"            :    "KC_DQUO",
    "KC_LEFT_ANGLE_BRACKET": "KC_LABK",
    "KC_LT"             :    "KC_LABK",
    "KC_RIGHT_ANGLE_BRACKET":"KC_RABK",
    "KC_QUESTION"       :    "KC_QUES"
}

keyplus_yaml_template = '''
# Generated by Q2K Keymap parser by 2Cas (c) 2018
# https://github.com/2Cas/Q2K
# This file is released into the public domain as per the CC0 Public Domain
# Dedication (http://creativecommons.org/publicdomain/zero/1.0/)
---

version: 0
report_mode: auto_nkro # options: auto_nkro, 6kro, nkro

devices:
  <KB_NAME> :
    id : 0
    layout : <LAYOUT_NAME>
    layout offset: 0
    scan_mode:
      mode: col_row # options: col_row, pins, none
      rows: <ROWS>
      cols: <COLS>
      # Maps how keys are physically wired, to how they appear visually
      matrix_map: [
        <MATRIX_MAP>
      ]
      # Debounce settings
      debounce:
         debounce_time_press: 5
         debounce_time_release: 5
         trigger_time_press: 1
         trigger_time_release: 1
         parasitic_discharge_delay_idle: 10.0
         parasitic_discharge_delay_debouncing: 10.0

layouts:
  <LAYOUT_NAME> :
    default_layer: 0
    layers: [
<LAYOUT>    ]
'''



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright 2018 2Cas
# Licensed under the MIT license (http://opensource.org/licenses/MIT)

import re, copy
import pyparsing as pp

from q2k.classes import *
from q2k.console import error_out, warning_out, note_out

def clean_split(line, char):

    newline = line.split(char)
    while '' in newline:
        newline.remove('')
        if not newline:
            newline = []
    return newline


def read_rules_mk(path):
    mcu_list = []
    try:
        with open(path, 'r') as f:
            data = str(f.read())
    except FileNotFoundError:
        warning_out(['Rules.mk not found in '+path,
            'Trying a different path...'])
        return
    
    EQUALS = pp.Suppress('=')
    mcu_tag = pp.Suppress(pp.Literal('MCU'))
    mcu_type = pp.Word(pp.alphanums+'_')

    mcu = mcu_tag + EQUALS + mcu_type('mcu')
    mcu.ignore('#'+pp.restOfLine)

    for tokens, start, end in mcu.scanString(data):
        mcu_list.append(tokens.mcu)
    
    return mcu_list


def read_config_header(data):

    data = str(data)
    matrix_pins = []
    matrix_row_pins = []
    matrix_col_pins = []
    
    LBRAC, RBRAC, COMMA = map(pp.Suppress, "{},")
    define_rows = pp.Suppress(pp.Literal('#define MATRIX_ROW_PINS'))
    define_cols = pp.Suppress(pp.Literal('#define MATRIX_COL_PINS'))
    pincode = pp.Word(pp.alphanums) | pp.Word(pp.nums)

    array = LBRAC + pp.ZeroOrMore(pincode + pp.Optional(COMMA)) + RBRAC
    matrix_rows = define_rows + LBRAC + pp.ZeroOrMore(pincode + pp.Optional(COMMA)) + RBRAC
    matrix_cols = define_cols + LBRAC + pp.ZeroOrMore(pincode + pp.Optional(COMMA)) + RBRAC

    for tokens, start, stop in matrix_rows.scanString(data):
        matrix_row_pins = list(tokens)
    for tokens, start, stop in matrix_cols.scanString(data):
        matrix_col_pins = list(tokens)   

    if matrix_row_pins and matrix_col_pins:
        matrix_pins.append(matrix_row_pins)
        matrix_pins.append(matrix_col_pins)

    return matrix_pins


def read_keymap(data):

    data = str(data)
    layer_list = []    
    LBRAC, RBRAC,EQ, COMMA = map(pp.Suppress,"{}=,")

    integer = pp.Word(pp.nums)
    keycode = pp.Word(pp.alphanums+'_'+'('+')')
    layer_string = pp.Word(pp.printables) | pp.Word(pp.nums)

    keycode_list = pp.Group(pp.ZeroOrMore(keycode + pp.Optional(COMMA)))
    row = LBRAC + keycode_list + RBRAC

    layern = layer_string('layer_name') + EQ
    km_layer_data = LBRAC + pp.OneOrMore(row + pp.Optional(COMMA)) + RBRAC
    km_layer = pp.Optional(layern) + km_layer_data('layer') + pp.Optional(COMMA)
    km_layer.ignore(pp.cppStyleComment+pp.pythonStyleComment)

    layer_list = []
    num_col = 0
    layer_index = -1
    for tokens, start, stop in km_layer.scanString(data):
        layer_index += 1
        if tokens.layer_name == '':
            curr_layer = keymap_layer(str(layer_index))
        else:
            curr_layer = keymap_layer(tokens.layer_name[1:-1]) 

        for row in tokens.layer:
            buff = ''
            for i, element in enumerate(row):
                if ')' in element and '(' not in element:
                    func = ''
                    func = row[i-1]+','+row[i] 
                    del row[i-1]
                    i -= 1
                    del row[i]
                    row.insert(i, func)

            num_col = len(row)
            curr_layer.add_keymap_item(list(row))
        
        curr_layer.set_matrix_cols(num_col)
        layer_list.append(curr_layer)
    """
    for layer in layer_list:
        print(layer.get_name()) 
        print(layer.get_keymap())
    """
    if not layer_list:

        error_out(['Parsed and found no keymap',
            'Failed to parse keymap file'])
        exit()
    else:
        return layer_list

def read_layout_header(path):

    template_list = []
    try:
        with open(path, 'r') as f:
            data = str(f.read())
    except FileNotFoundError:
        warning_out(['Layout header not found in '+path,
            'Trying a different path...'])
        return
    
    LPAREN, RPAREN, LBRAC, RBRAC, COMMA = map(pp.Suppress, "(){},")
    BSLASH = pp.Suppress(pp.Literal('\\'))

    define = pp.Suppress(pp.Literal('#define'))
    name = pp.Word(pp.alphanums+'_')
    macrovar = pp.Word(pp.alphanums+'_')
    layout_row = pp.Group(pp.OneOrMore(macrovar + pp.Optional(COMMA)) + pp.ZeroOrMore(BSLASH))
    layout =  LPAREN + pp.ZeroOrMore(BSLASH) + pp.OneOrMore(layout_row) + RPAREN

    matrix_row = pp.ZeroOrMore(BSLASH) + pp.Optional(LBRAC) +  pp.ZeroOrMore(BSLASH) + pp.OneOrMore(macrovar + pp.Optional(COMMA)) + pp.ZeroOrMore(RBRAC) + pp.Optional(COMMA) + pp.ZeroOrMore(BSLASH)
    matrix = pp.ZeroOrMore(BSLASH) + LBRAC +  pp.ZeroOrMore(BSLASH) + pp.OneOrMore(matrix_row) +  pp.ZeroOrMore(BSLASH) + RBRAC +  pp.ZeroOrMore(BSLASH)

    header = define + name('name') + pp.ZeroOrMore(BSLASH) + layout('layout') + pp.ZeroOrMore(BSLASH) + matrix('array')
    header.ignore(pp.cStyleComment)

    for tokens, start, end in header.scanString(data):
        curr_template = layout_template(tokens.name)

        for row in tokens.layout:
            layout_row = list(row)
            curr_template.add_layout_line(layout_row)
       
        array = list(tokens.array)
        for i, element in enumerate(array):
            array[i] = re.sub('^[^##]*##', '', element) 

        curr_template.set_array(array)
        template_list.append(curr_template)
    
    # Translate macro k vars to array indices
    for x, t in enumerate(template_list):
        layout = t.get_layout()
        array = t.get_array()
        for i, row in enumerate(layout):
            for j, col in enumerate(row):
                try:
                    layout[i][j] = array.index(col)
                except ValueError:
                    warning_out(['Missing macro variable ['+col+'] in '+path])
                    # Try to recover using a different keycode array
                    for y, t2 in enumerate(template_list):                    
                        array2 = t2.get_array()
                        if len(array2) != len(array) or x == y:
                            continue
                        try:
                            index = array2.index(col)
                            layout[i][j] = index
                            array[index] = col                            
                            warning_out(['Macro variable recovery succeeded'])
                            break
                        except ValueError:
                            continue
                    if layout[i][j] == col:
                        error_out(['Array key recovery failed',
                            'Missing macro variable in: '+path])
                        exit()
        t.set_layout(layout)

    """
    for k in template_list:
        layout = k.get_layout()
        array = k.get_array()
        print(k.get_name())
        for line in layout:
            print(line)
        print('array')
        print(array)
    """
    return template_list



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright 2018 2Cas
# Licensed under the MIT license (http://opensource.org/licenses/MIT)

# A class for listing keyboards and QMK folder structure associated with keyboards (can be dumped and loaded to yaml)

class kb_info:

    def __init__(self, n=''):
        self._kbname = n
        self._revL = []
        self._keymapL = []       
        self._klibs = []

        self.keymap_name = ''
        self.rev = ''
        self.rev_info = []

    def set_rev(self, rev):
        self.rev = rev

    def set_keymap(self, keymap):
        self.keymap_name = keymap
        
    def set_libs(self, lib):
        self._klibs = lib

    def add_rev_list(self, rev, flag=False):
        if flag == False:
            self._revL.append(rev)
        revObj = rev_info(rev, flag)
        self.rev_info.append(revObj)

    def add_keymap_list(self, km):
        self._keymapL.append(km)

    def add_lib(self, lib):
        self._klibs.append(lib)
        print(lib)
        
    def get_name(self):
        return self._kbname

    def get_keymap(self):
        return self.keymap_name

    def get_libs(self):
        return self._klibs

    def get_rev(self):
        return self.rev

    def get_rev_info(self, rev):
        for r in self.rev_info:
           if r._name == rev:
               return r
           if r._name == 'n/a':
               return r
    def get_keymap_list(self):
        return self._keymapL

    def get_rev_list(self):
        """
        rev_list = []        
        for r in self.rev_info:
            if r.isDefault == False:
                rev_list.append(r._name)
        return rev_list
        """
        return self._revL

# A class for listing revision info and 
class rev_info:
    def __init__ (self, n='', flag=False):
        self._name = n
        self.isDefault = flag

        self.build_m_row_pins = []
        self.build_m_col_pins = []
        self.build_layout = []
        self.build_templates = []

        self.mcuL = []
        #self.output_keymap = ''

    def set_mcu_list(self, mcu):
        self.mcuL = mcu

    def set_layout(self, lot_list):
        self.build_layout = lot_list

    def set_templates(self, temp_list):
        self.build_templates = temp_list

    def set_matrix_pins(self, r, c):
        self.build_m_row_pins = r
        self.build_m_col_pins = c

    def set_output_keymap(self, filepath):
        self.output_keymap = filepath

    def get_rev_name(self):
        return self._name

    def get_mcu_list(self):
        return self.mcuL

    def get_layout(self):
        return self.build_layout

    def get_templates(self):
        return self.build_templates

    def get_matrix_pins(self):
        matrix = []
        matrix.append(self.build_m_row_pins)
        matrix.append(self.build_m_col_pins)
        return matrix

    #def get_output_keymap(self):
        #return self.output_keymap


# A class for linking matrices in <keyboard>.h with (preprocessed) keymap.c layouts
class layout_template:

    def __init__(self, n=''):
        self.layout_name = n
        self.layout = []
        self.array = []

    def set_array(self, a):
        self.array = a

    def set_layout(self, l):
        self.layout = l

    def add_layout_line(self, l):
        self.layout.append(l)

    def add_array_item(self, item):
        self.array = self.array+item

    def get_name(self):
        return self.layout_name

    def get_array(self):
        return self.array

    def get_layout(self):
        return self.layout

# A class for storing keymap layers
class keymap_layer:

    def __init__(self, n=''):
        self.keymap = []
        self.layer_name = n

        self.layout = []
        self.template = []
        self.matrix_cols = 0

    def set_keymap(self, k):
        self.keymap = k

    def set_matrix_cols(self, c):
        self.matrix_cols = c
    
    def set_template(self, m):
        self.template = m

    def set_layout(self, l):
        self.layout = l

    def add_keymap_item(self, l):
        self.keymap = self.keymap+l

    def get_name(self):
        return self.layer_name

    def get_keymap(self):
        return self.keymap

    def get_layout(self):
        return self.layout

    def get_template(self):
        return self.template

    def get_matrix_cols(self):
        return self.matrix_cols



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright 2018 2Cas
# Licensed under the MIT license (http://opensource.org/licenses/MIT)


from q2k.classes import *
from q2k.globals import KBD_LIST, KB_INFO, MCU_COMPAT

from q2k.parser import *
from q2k.convert import *
from q2k.cpp import *
from q2k.yaml import *
from q2k.console import error_out, warning_out, note_out
from q2k.json import *

def print_keyboard_list():

    printkb = []
    for kbc in KBD_LIST:
        printkb.append(kbc.get_name())
    print(printkb)    


def print_keymap_list(kb):

    printkm = []
    for kbc in KBD_LIST:
        if kbc.get_name() == kb:
            for km in kbc.get_keymap_list():
                printkm.append(km)
    print(printkm)


def print_revision_list(kb):

    printr = []
    for kbc in KBD_LIST:
        if kbc.get_name() == kb:
            for r in kbc.get_rev_list():
                printr.append(r)
    if len(printr) > 0:
        print(printr)
    else:
        print("None")


def search_keyboard_list(s):

    printkb = []
    for kbc in KBD_LIST:
        kbc_name = kbc.get_name()
        if kbc_name.find(s) != -1:
            printkb.append(kbc_name)
    print(printkb)
    

def check_parse_argv(kb, km='default', rev=''):

    for kbc in KBD_LIST:
        # Check keyboard
        if kbc.get_name() == kb:
            # Check Keymap
            if km not in kbc.get_keymap_list():
                warning_out(['Invalid Keymap - '+km,
                    'Valid Keymaps:'])
                print_keymap_list(kb)
                exit()
            # Check Revision
            # Case 1: Have Revisions
            if len(kbc.get_rev_list()) != 0 and rev not in kbc.get_rev_list():
                error_out(['Invalid Revision - '+rev,
                    'Valid Revisions:'])
                print_revision_list(kb)
                exit()
            # Case 2: No Revisions
            elif len(kbc.get_rev_list()) == 0 and rev != '':
                error_out(['Invalid Revision - '+rev,
                    'Valid Revisions:'])
                print_revision_list(kb)
                exit()
            # If KB matches, and KM and REV do not conflict, set values and return kb_info class
            kbc.set_rev(rev)
            kbc.set_keymap(km)
            return kbc
        else:
            continue
    
    # If we finish loop with no matches, keyboard name must be invalid
    error_out(['Invalid Keyboard Name - '+kb,
        'Valid Names:'])
    print_keyboard_list()
    exit()

def init_cache_info(kb_info_yaml):

    if os.path.isfile(kb_info_yaml):
        global KBD_LIST
        try:
            with open(kb_info_yaml, 'r') as f:
                KBD_LIST = yaml.load(f)
                note_out(['Using cached kb_info.yaml @ '+KB_INFO])

        except FileNotFoundError:
            warning_out(['Failed to load kb_info.yaml',
                'Generating new kb_info.yaml...'])
            write_info(kb_info_yaml)

        except ImportError:
            warning_out(['Failed to load kb_info.yaml',
                'Generating new kb_info.yaml...'])
            write_info(kb_info_yaml)
    else:
            write_info(kb_info_yaml)
            note_out(['New kb_info.yaml successfully generated', 'Location: '+KB_INFO])


def find_rules_mk(kbc):
    rev = kbc.get_rev()
    revObj = kbc.get_rev_info(rev)
    kblibs = list(kbc.get_libs())
    if rev != '':
        kblibs.append(rev)

    qdir = QMK_DIR +'keyboards/'
    folders = []
    path = ''

    for kbl in kblibs:
        path += kbl+'/'
        folders.append(path)

    for kbl in reversed(folders):
        rules_mk = 'rules.mk'

        path = qdir+kbl+rules_mk
        mcu_list = read_rules_mk(path)
        if mcu_list:
            revObj.set_mcu_list(mcu_list)
            return mcu_list
        else:
            continue

    warning_out(['Rules.mk not found for '+kbc.get_name()])
    return

def check_mcu_list(kbc):
    kb = kbc.get_name()
    rev = kbc.get_rev()
    revObj = kbc.get_rev_info(rev)
    mcu_list = revObj.get_mcu_list()

    for mcu in mcu_list:
        if mcu in MCU_COMPAT:
            continue
        else:
            warning_out(
                ['Possible MCU incompatability detected', 
                'MCU type: '+mcu+' in '+kb+'/'+rev+' rules.mk', 
                'Currently, keyplus supports only boards with the following microcontrollers:',
                str(MCU_COMPAT),
                'If your board has a MCU on this list then ignore this warning as a false positive',
                'Else layout files produced may not work with keyplus until your board\'s mcu is supported',
                'Press [ENTER] to continue'])
            input()
            return
    note_out(['No MCU incompatability detected'])


def find_config_header(kbc):
    rev = kbc.get_rev()
    revObj = kbc.get_rev_info(rev)
    kblibs = list(kbc.get_libs())
    if rev != '':
        kblibs.append(rev)

    qdir = QMK_DIR +'keyboards/'
    folders = []
    path = ''

    for kbl in kblibs:
        path += kbl+'/'
        folders.append(path)

    for kbl in reversed(folders):
        config_h = 'config.h'

        path = qdir+kbl+config_h
        data = preproc_header(kbc, path)
        matrix_pins = read_config_header(data)
        if matrix_pins:
            revObj.set_matrix_pins(matrix_pins[0], matrix_pins[1])
            note_out(['Matrix pinout data found @ '+path])
            return matrix_pins
        else:
            continue

    warning_out(['Config.h header not found for '+kbc.get_name(),
        'Matrix row/col pins must be provided manually!']) 
    return


def preproc_read_keymap(kbc):
    rev = kbc.get_rev()
    revObj = kbc.get_rev_info(rev)

    data = preproc_keymap(kbc)
    layer_list = read_keymap(data)
    revObj.set_layout(layer_list)

    return layer_list


def find_layout_header(kbc):

    rev = kbc.get_rev()
    revObj = kbc.get_rev_info(rev)
    # qmk/keyboards/
    kblibs = list(kbc.get_libs())
    if rev != '':
        kblibs.append(rev)
    qdir = QMK_DIR +'keyboards/'

    folders = []
    path = ''
    for kbl in kblibs:
        path += kbl+'/'
        folders.append(path)

    for kbl in reversed(folders):
        kbl_f = kbl.split('/')
        kb_h = kbl_f[-2]        
        kb_h += '.h'

        path = qdir+kbl+kb_h
        #data = preproc_header(kbc, path)
        #matrix_layout = read_layout_header(data)
        matrix_layout = read_layout_header(path)
        if matrix_layout:
            revObj.set_templates(matrix_layout)
            note_out(['Layout header found @ '+path])
            return matrix_layout
        else:
            continue
    warning_out(['Keyboard layout header not found for '+kbc.get_name(), 'Reverting to basic layout...']) 
    return

# Hook in for main q2k binary
def main():   
    # Init kb_info file from cache
    init_cache_info(KB_INFO)
    # Read ARGV input from terminaL
    parser = argparse.ArgumentParser(description='Convert AVR C based QMK keymap and matrix files to YAML Keyplus format')
    parser.add_argument('keyboard', metavar='KEYBOARD', nargs='?', default='', help='The name of the keyboard whose keymap you wish to convert')
    parser.add_argument('-m', '--keymap', metavar='keymap', dest='keymap', default='default', help='The keymap folder to reference - default is /default/')
    parser.add_argument('-r', '--rev', metavar='ver',dest='rev', default='', help='Revision of layout - default is n/a')
    parser.add_argument('-d', '--dump', dest='dumpyaml', action='store_true', help='Append results to kb_info.yaml. (For debugging, May cause performance penalty)')
    parser.add_argument('-L', '--list', dest='listkeyb', action='store_true',help='List all valid KEYBOARD inputs')
    parser.add_argument('-M', '--keymaps', dest='listkeym', action='store_true',help='List all valid KEYMAPS for the current keyboard')
    parser.add_argument('-R', '--revlist', dest='listkeyr', action='store_true',help='List all valid REVISIONS for the current keyboard')
    parser.add_argument('-S', '--search', metavar='string', dest='searchkeyb', help='Search valid KEYBOARD inputs')
    parser.add_argument('-p', '--print', dest='presult', action='store_true',help='Print result of keymap conversion to terminal')
    parser.add_argument('-t', '--template', metavar='layout', type=int, default=-1, dest = 'choosemap', help= 'Select keymap template index')
    args = parser.parse_args()

    if args.listkeyb:
        print_keyboard_list()
        sys.exit()

    if args.listkeym:
        note_out(['Listing keymaps for '+args.keyboard+'...'])
        print_keymap_list(args.keyboard)
        sys.exit()

    if args.listkeyr:
        note_out(['Listing revisions for '+args.keyboard+'...'])
        print_revision_list(args.keyboard)
        sys.exit()

    if args.searchkeyb:
        note_out(['Searching...'])
        search_keyboard_list(args.searchkeyb)
        sys.exit()

    current_kbc = check_parse_argv(args.keyboard, args.keymap, args.rev)        # Check the cmd line arguments
    mcu_list = find_rules_mk(current_kbc)                                       # Find and check MCU type
    check_mcu_list(current_kbc)
    find_config_header(current_kbc)                                             # Pass config.h through CPP for matrix pinout
    km_layers = preproc_read_keymap(current_kbc)                                # Check cache/run preprocessor for keymap.c
    km_template = find_layout_header(current_kbc)                               # Find layout templates in <keyboard>.h
    km_layers = convert_keymap(km_layers)                                       # Convert extracted keymap.c data to keyplus format
    if not km_template:
        km_template = build_layout_from_keymap(km_layers)
    merge_layout_template(km_layers, km_template, args.choosemap)               # Merge layout templates + arrays from <keyboard>.h with matrix from keymap.c
    note_out(['Conversion succeeded, piping output'])

    if args.dumpyaml:
        dump_info(KB_INFO)

    if args.presult:
        create_keyplus_yaml(current_kbc, True)
    else:
        create_keyplus_yaml(current_kbc)

# hook in for q2kb binary
def kbfirmware():   
    # Init kb_info file from cache
    init_cache_info(KB_INFO)
    # Read ARGV input from terminaL
    parser = argparse.ArgumentParser(description='Convert AVR C based QMK keymap and matrix files to YAML Keyplus format')
    parser.add_argument('keyboard', metavar='KEYBOARD', nargs='?', default='', help='The name of the keyboard whose keymap you wish to convert')
    parser.add_argument('-m', '--keymap', metavar='keymap', dest='keymap', default='default', help='The keymap folder to reference - default is /default/')
    parser.add_argument('-r', '--rev', metavar='ver',dest='rev', default='', help='Revision of layout - default is n/a')
    parser.add_argument('-d', '--dump', dest='dumpyaml', action='store_true', help='Append results to kb_info.yaml. (For debugging, May cause performance penalty)')
    parser.add_argument('-L', '--list', dest='listkeyb', action='store_true',help='List all valid KEYBOARD inputs')
    parser.add_argument('-M', '--keymaps', dest='listkeym', action='store_true',help='List all valid KEYMAPS for the current keyboard')
    parser.add_argument('-R', '--revlist', dest='listkeyr', action='store_true',help='List all valid REVISIONS for the current keyboard')
    parser.add_argument('-S', '--search', metavar='string', dest='searchkeyb', help='Search valid KEYBOARD inputs')
    parser.add_argument('-p', '--print', dest='presult', action='store_true',help='Print result of keymap conversion to terminal')
    parser.add_argument('-t', '--template', metavar='layout', type=int, default=-1, dest = 'choosemap', help= 'Select keymap template index')
    args = parser.parse_args()

    if args.listkeyb:
        print_keyboard_list()
        sys.exit()

    if args.listkeym:
        note_out(['Listing keymaps for '+args.keyboard+'...'])
        print_keymap_list(args.keyboard)
        sys.exit()

    if args.listkeyr:
        note_out(['Listing revisions for '+args.keyboard+'...'])
        print_revision_list(args.keyboard)
        sys.exit()

    if args.searchkeyb:
        note_out(['Searching...'])
        search_keyboard_list(args.searchkeyb)
        sys.exit()
    
    current_kbc = check_parse_argv(args.keyboard, args.keymap, args.rev)        # Check the cmd line arguments
    mcu_list = find_rules_mk(current_kbc)                                       # Find and check MCU type
    check_mcu_list(current_kbc)
    find_config_header(current_kbc)                                             # Pass config.h through CPP for matrix pinout
    km_layers = preproc_read_keymap(current_kbc)                                # Check cache/run preprocessor for keymap.c
    km_template = find_layout_header(current_kbc)                               # Find layout templates in <keyboard>.h
    if not km_template:
        km_template = build_layout_from_keymap(km_layers)
    merge_layout_template(km_layers, km_template, args.choosemap)               # Merge layout templates + arrays from <keyboard>.h with matrix from keymap.c
    note_out(['Merge succeeded, piping output'])

    if args.dumpyaml:
        dump_info(KB_INFO)

    if args.presult:
        create_kbfirmware_json(current_kbc, True)
    else:
        create_kbfirmware_json(current_kbc)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright 2018 2Cas
# Licensed under the MIT license (http://opensource.org/licenses/MIT)

import sys
from termcolor import colored, cprint

def error_out(info):

    error = colored('❌ ERROR:', 'red', attrs=['reverse', 'bold'])
    e_bullet = colored('•', 'red', attrs=['bold'])
    
    for info, line in enumerate(info):
        if not info:
            line = colored(line, 'red')
            print(error + ' ' +line, file=sys.stderr)
        else:
            print(e_bullet + ' '+line,file=sys.stderr)


def bad_kc_out(kc_type, code):

    bad_kc = colored('❌ Invalid '+kc_type+':', 'cyan')
    #code = colored(code, 'cyan')
    print(bad_kc+' '+code, file=sys.stderr)


def warning_out(info):

    warning = colored('▲ WARNING:', 'yellow', attrs=['bold'])
    w_bullet = colored('•', 'yellow', attrs=['bold'])
    
    for info, line in enumerate(info):
        if not info:
            line = colored(line, 'yellow')
            print(warning + ' ' +line, file=sys.stderr)
        else:
            print(w_bullet + ' '+line, file=sys.stderr)


def note_out(info):

    note = colored('✔', 'green', attrs=['bold'])
    n_bullet = colored('•', 'green', attrs=['bold'])
    
    for info, line in enumerate(info):
        if not info:
            #line = colored(line, 'green')
            print(note + ' ' +line)
        else:
            print(n_bullet + ' '+line)


